from typing import Dict, List, Set, Tuple, Type, Union

import abstracts


class Implementer(type):
    """Metaclass for implementers of an Abstract interface.

    Any ``Abstraction`` classes that are listed in ``__implements__`` for the
    class are added as bases (ie added to the class's inheritance).

    Any docs for methods are copied from the interface method to the
    implementation if the implementation method has no docs of its own.

    For example:

    .. code-block:: python

        from tools.base.abstract import Abstraction, Implementer


        class AFoo(metaclass=Abstraction):

            @abstractmethod
            def do_something(self):
                ""\"Do something""\"
                raise NotImplementedError

        class Foo(metaclass=Implementer)
            __implements__ = (AFoo, )

            def do_something(self):
                return "DONE"

    Given the above, you should see that instantiating `Foo`:

    .. code-block:: pycon

        >>> isinstance(Foo(), AFoo)
        True
        >>> Foo().do_something.__doc__
        'Do something'
    """

    @classmethod
    def abstract_info(
            cls,
            abstract: "abstracts.Abstraction") -> Tuple[
                str, Union[str, None], List[str]]:
        """Information for a specific abstract implementation class.

        For given abstract class, returns:

        - qualified class name
        - class docstring
        - abstract methods
        """
        if not isinstance(abstract, abstracts.Interface):
            raise TypeError(
                "Implementers can only implement subclasses of "
                f"`abstracts.Abstraction`, unrecognized: '{abstract}'")
        methods: List[str] = []
        for method in getattr(abstract, "__abstractmethods__", []):
            methods.append(method)
        return (
            f"{abstract.__module__}.{abstract.__name__}",
            abstract.__doc__,
            methods)

    @classmethod
    def add_docs(cls, clsdict: Dict, klass: "Implementer") -> None:
        """Add docs to the implementation class.

        If the implementation class has no docstring, then a docstring is
        generated with the format:

        .. code-block::

            Implements: foo.bar.ABaz
            An implementer of the ABaz protocol...

            Implements: foo.bar.AOtherBaz
            An implementer of the AOtherBaz protocol...

        For each of the methods that are marked abstract in any of the abstract
        classes, if the method in the implementation class has no docstring the
        docstring is resolved from the abstract methods, using standard mro.
        """
        abstract_docs, abstract_methods = cls.implementation_info(clsdict)
        if not klass.__doc__:
            klass.__doc__ = "\n".join(
                f"Implements: {k}\n{v}\n" for k, v in abstract_docs.items())
        for abstract_method, abstract_klass in abstract_methods.items():
            method = getattr(klass, abstract_method, None)
            if not method:
                # this will not instantiate, so bail now
                return
            # Only set the doc for the method if its not already set.
            # `@classmethod` `__doc__`s are immutable, so skip them.
            if not method.__doc__ and not hasattr(method, "__self__"):
                method.__doc__ = getattr(
                    abstract_klass, abstract_method).__doc__

    @classmethod
    def add_interfaces(
            cls,
            ifaces: Tuple,
            klass: "Implementer") -> None:
        for iface in ifaces:
            if issubclass(klass, iface):
                continue
            missing = iface.__abstractmethods__ - set(dir(klass))
            if missing:
                raise TypeError(
                    "Not all methods for interface "
                    f"{iface} provided: missing {sorted(missing)}")
            iface.register(klass)

    @classmethod
    def check_interface(
            cls,
            clsdict: dict) -> None:
        extra = (
            cls.get_class_attrs(clsdict)
            - cls.get_interface_methods(clsdict))
        if extra:
            raise TypeError(
                "Interfaces can only contain methods decorated with "
                f"`@interfacemethod`: got {extra}")

    @classmethod
    def get_bases(
            cls,
            bases: Tuple[Type, ...],
            clsdict: Dict) -> Tuple[Type["abstracts.Abstraction"], ...]:
        """Returns a tuple of base classes, with `__implements__` classes
        included."""
        return (
            bases
            + tuple(
                x for x in clsdict["__implements__"]
                if x not in bases
                and isinstance(x, abstracts.Abstraction)))

    @classmethod
    def get_class_attrs(
            cls,
            clsdict: dict) -> Set[str]:
        return set(
            x for x in clsdict
            if not x.startswith("__"))

    @classmethod
    def get_interface_methods(
            cls,
            clsdict: Dict) -> Set[str]:
        return set(
            k for k, v in clsdict.items()
            if (isinstance(v, property)
                and getattr(v.fget, "__isinterfacemethod__", False))
            or getattr(v, "__isinterfacemethod__", False)
            or (isinstance(v, classmethod)
                and getattr(v.__func__, "__isinterfacemethod__", False)))

    @classmethod
    def get_interfaces(
            cls,
            bases: Tuple[Type, ...],
            clsdict: Dict) -> Tuple[Type["abstracts.Interface"], ...]:
        return tuple(
            x for x in clsdict["__implements__"]
            if (x not in bases and cls.is_interface(x)))

    @classmethod
    def implementation_info(
            cls,
            clsdict: Dict) -> Tuple[Dict[str, str], Dict[str, Type]]:
        """Returns 2 dictionaries.

        - abstract_docs: abstract docs for all abstract classes
        - abstract_methods: resolved abstract methods -> abstract class
        """
        abstract_docs: Dict[str, str] = {}
        abstract_methods: Dict[str, Type] = {}
        for abstract in reversed(clsdict["__implements__"]):
            name, docs, methods = cls.abstract_info(abstract)
            for method in methods:
                abstract_methods[method] = abstract
            if docs:
                abstract_docs[name] = docs
        return abstract_docs, abstract_methods

    @classmethod
    def is_interface(
            cls,
            klass: Union["abstracts.Interface", "Implementer"]) -> bool:
        return (
            isinstance(klass, abstracts.Interface)
            and not isinstance(klass, abstracts.Abstraction))

    def __new__(
            cls,
            clsname: str,
            bases: Tuple[Type, ...],
            clsdict: Dict) -> "Implementer":
        """Create a new Implementer class."""
        if "__implements__" not in clsdict:
            klass = super().__new__(cls, clsname, bases, clsdict)
            if cls.is_interface(klass):
                cls.check_interface(clsdict)
            return klass
        ifaces = cls.get_interfaces(bases, clsdict)
        klass = super().__new__(
            cls, clsname, cls.get_bases(bases, clsdict), clsdict)
        cls.add_docs(clsdict, klass)
        cls.add_interfaces(ifaces, klass)
        return klass



from abc import abstractmethod
from functools import wraps
from typing import Any

from abstracts.implements import Implementer


def implementer(implements):
    """Decorator for implementers.

    Assuming you have an abstract class `AFoo` which has a `metaclass` of
    type `Abstraction`, it can be used as follows:

    ```
    from tools.base import abstract

    @abstract.implementer(AFoo)
    class Foo:

        def bar(self):
            return "BAZ"

    ```
    """
    if not isinstance(implements, (tuple, list, set)):
        implements = (implements,)

    def wrapper(klass: Any) -> Implementer:
        # This is a v annoying workaround for mypy, see:
        #   https://github.com/python/mypy/issues/9183
        _klass: Any = klass

        class Implementation(_klass, metaclass=Implementer):
            __implements__ = implements
            __doc__ = _klass.__doc__

        Implementation.__module__ = klass.__module__
        Implementation.__qualname__ = klass.__name__
        Implementation.__name__ = klass.__name__
        return Implementation

    return wrapper


def interfacemethod(fun):

    @wraps(fun)
    @abstractmethod
    def wrapped(*args, **kwargs):
        raise NotImplementedError

    wrapped.__isinterfacemethod__ = True
    return wrapped




# DO NOT EDIT THIS FILE -- AUTOGENERATED BY PANTS
# Target: py/abstracts:package

from setuptools import setup

setup(**{
    'author': 'Ryan Northey',
    'author_email': 'ryan@synca.io',
    'description': 'Abstract class and interface definitions',
    'install_requires': (
    ),
    'license': 'Apache Software License 2.0',
    'long_description': """
abstracts
=========

Abstract class and interface definitions.

Create an ``abstract.Abstraction``
----------------------------------

An ``Abstraction`` is a ``metaclass`` for defining abstract classes.

Let's define an abstract ``AFoo`` class and give it an abstract ``do_foo``
method.

Like any python class, an ``Abstraction`` can have any name, but it may
be helpful to distinguish abstract classes from others by prefixing their
name with ``A``.

.. code-block:: pycon

   >>> import abc
   >>> import abstracts

   >>> class AFoo(metaclass=abstracts.Abstraction):
   ...
   ...     @abc.abstractmethod
   ...     def do_foo(self):
   ...         raise NotImplementedError

Abstract classes **cannot** be instantiated directly.

.. code-block:: pycon

   >>> AFoo()
   Traceback (most recent call last):
   ...
   TypeError: Can't instantiate abstract class AFoo with abstract method... do_foo

Create an ``implementer`` for an ``abstract.Abstraction``
---------------------------------------------------------

In order to make use of ``AFoo``, we need to create an implementer for it.

.. code-block:: pycon

   >>> @abstracts.implementer(AFoo)
   ... class Foo:
   ...     pass

The implementer **must** implement all of the abstract methods,
defined by its abstract classes.

.. code-block:: pycon

   >>> Foo()
   Traceback (most recent call last):
   ...
   TypeError: Can't instantiate abstract class Foo with abstract method... do_foo

   >>> @abstracts.implementer(AFoo)
   ... class Foo2:
   ...
   ...     def do_foo(self):
   ...         return "DID FOO"

   >>> Foo2()
   <__main__.Foo2 object at ...>

An implementer inherits from its ``Abstractions``
-------------------------------------------------

An ``implementer`` class is a subclass of its ``Abstraction``.

.. code-block:: pycon

   >>> issubclass(Foo2, AFoo)
   True

Likewise an instance of an implementer is an instance of its ``Abstraction``

.. code-block:: pycon

   >>> isinstance(Foo2(), AFoo)
   True

The ``Abstraction`` class can be seen in the class ``bases``, and the
methods of the ``Abstraction`` can be invoked by the implementer.

.. code-block:: pycon

   >>> import inspect
   >>> AFoo in inspect.getmro(Foo2)
   True

Create an ``implementer`` that implements multiple ``Abstraction`` s.
---------------------------------------------------------------------

An implementer can implement multiple abstractions.

Let's create a second abstraction.

.. code-block:: pycon

   >>> class ABar(metaclass=abstracts.Abstraction):
   ...
   ...     @abc.abstractmethod
   ...     def do_bar(self):
   ...         raise NotImplementedError

And now we can create an implementer that implememts both the ``AFoo`` and ``ABar``
``Abstraction`` s.

.. code-block:: pycon

   >>> @abstracts.implementer((AFoo, ABar))
   ... class FooBar:
   ...
   ...     def do_foo(self):
   ...         return "DID FOO"
   ...
   ...     def do_bar(self):
   ...         return "DID BAR"

   >>> FooBar()
   <__main__.FooBar object at ...>

Defining abstract properties
----------------------------

Properties can be defined in an abstract class, and just like with normal
methods, they must be implemented by any implementers.

.. code-block:: pycon

   >>> class AMover(metaclass=abstracts.Abstraction):
   ...
   ...     @property
   ...     @abc.abstractmethod
   ...     def speed(self):
   ...         return 5
   ...
   ...     @property
   ...     @abc.abstractmethod
   ...     def direction(self):
   ...         return "forwards"

Calling ``super()`` on an ``abstractmethod``
--------------------------------------------

Just like with pythons "Abstract Base Classes" you can call ``super()``
in an ``abstractmethod``, to invoke an abstract implementation.

.. code-block:: pycon

   >>> @abstracts.implementer(AMover)
   ... class Mover:
   ...
   ...     @property
   ...     def direction(self):
   ...         return "backwards"
   ...
   ...     @property
   ...     def speed(self):
   ...         return super().speed

This custom implementation of ``AMover`` **must** implement both ``speed`` and
``direction``, even if its implementation invokes the abstract implementation.

In this case it uses the default/abstract implementation of ``speed`` while providing
its own implementation of ``direction``.

.. code-block:: pycon

   >>> mover = Mover()
   >>> mover
   <__main__.Mover object at ...>

   >>> mover.speed
   5
   >>> mover.direction
   'backwards'

Defining an ``abstracts.Interface`` class
-----------------------------------------

An ``Interface`` is much like an ``Abstraction``, but with a few differences.

An ``Interface`` can only define methods with the ``@interfacemethod`` decorator.

It cannot define normal methods or methods with the ``@abstractmethod``, only methods
with ``@interfacemethod``.

An ``@interfacemethod`` if invoked will always raise an ``NotImplementedError``, and
therefore cannot be used as an abstract implementation.

Lets add an ``Interface`` class that we can use.

In the way that it may be helpful to distinguish an ``Abstraction`` from other
types of classes, it may be also useful to distinguish an ``Interface`` by
using an ``I`` prefix when naming them.

.. code-block:: pycon

   >>> class IGeared(metaclass=abstracts.Interface):
   ...
   ...     @property
   ...     @abstracts.interfacemethod
   ...     def number_of_gears(self):
   ...         # Raising an error is ~superfluous as the decorator will raise
   ...         # anyway if the method is invoked.
   ...         raise NotImplementedError

Implementing an ``Interface``
-----------------------------

Just like with an ``Abstraction``, an ``Interface`` can be implemented using
the ``@implementer`` decorator.

An implementer, can implement a combination of ``Abstractions`` and
``Interfaces``.

.. code-block:: pycon

   >>> @abstracts.implementer((AMover, IGeared))
   ... class Bicycle:
   ...
   ...     @property
   ...     def direction(self):
   ...         return super().direction
   ...
   ...     @property
   ...     def speed(self):
   ...         return super().speed
   ...
   ...     @property
   ...     def number_of_gears(self):
   ...         return 7

   >>> Bicycle().number_of_gears
   7

An implementer does **not** inherit from its ``Interfaces``
-----------------------------------------------------------

An ``implementer`` class is a subclass of its ``Interfaces``.

.. code-block:: pycon

   >>> issubclass(Bicycle, AMover)
   True
   >>> issubclass(Bicycle, IGeared)
   True

Likewise an instance of an implementer is an instance of its ``Interfaces``

.. code-block:: pycon

   >>> isinstance(Bicycle(), AMover)
   True
   >>> isinstance(Bicycle(), IGeared)
   True

Unlike with ``Abstractions`` it does **not** however, inherit from its ``Interfaces``.

.. code-block:: pycon

   >>> AMover in inspect.getmro(Bicycle)
   True

   >>> IGeared in inspect.getmro(Bicycle)
   False

``@interfacemethods`` can never be invoked
------------------------------------------

The key thing to remember is that you cannot call ``super()`` on any
``@interfacemethod``, or directly invoke it.

If it was defined as part of an ``Interface`` you will receive an
``AttributeError``, as the implementation does not inherit directly from the
interface.

.. code-block:: pycon

   >>> @abstracts.implementer((AMover, IGeared))
   ... class BrokenBicycle:
   ...
   ...     @property
   ...     def direction(self):
   ...         return super().direction
   ...
   ...     @property
   ...     def speed(self):
   ...         return super().speed
   ...
   ...     @property
   ...     def number_of_gears(self):
   ...         return super().number_of_gears

   >>> BrokenBicycle().number_of_gears
   Traceback (most recent call last):
   ...
   AttributeError: 'super' object has no attribute 'number_of_gears'

.. warning::

   Misuse of this class can have `unintended consequences <https://www.dailymotion.com/video/x2howud>`_

If you invoke ``super()`` on an ``@interfacemethod`` defined as part of an
``Abstraction`` it will raise ``NotImplementedError``.

As an ``Interface`` can only hold this type of method, you can never invoke
any of its methods. Doing so directly will raising a ``NotImplementedError``.

.. code-block:: pycon

   >>> IGeared.number_of_gears.__get__(Bicycle())
   Traceback (most recent call last):
   ...
   NotImplementedError

Combining ``@abstractmethod`` and ``@interfacemethod`` in an ``Abstraction``
----------------------------------------------------------------------------

As ``Interfaces`` are "pure", they cannot use ``@abstractmethod`` or contain any implementation.

An ``Abstraction`` on the other hand can combine both.

Lets create a pure ``Interface`` that represents a "shed".

.. code-block:: pycon

   >>> class IShed(metaclass=abstracts.Interface):
   ...
   ...     @property
   ...     @abstracts.interfacemethod
   ...     def size(self):
   ...         raise NotImplementedError

We can use this interface to create an ``ABikeShed`` ``Abstraction``

.. code-block:: pycon

   >>> class ABikeShed(IShed, metaclass=abstracts.Abstraction):
   ...
   ...     @property
   ...     @abstracts.interfacemethod
   ...     def max_bike_size(self):
   ...         raise NotImplementedError
   ...
   ...     @abc.abstractmethod
   ...     def get_capacity(self):
   ...         return int(self.size / self.max_bike_size)

We can now create an implementation.

It will need to define both the ``size`` and the ``max_bike_size``,
as these are ``interfacemethods``.

It can, however, make use of the abstract implementation of ``get_capacity``,
even if it must be defined.

.. code-block:: pycon

   >>> @abstracts.implementer(ABikeShed)
   ... class BikeShed:
   ...
   ...     @property
   ...     def max_bike_size(self):
   ...         return 7
   ...
   ...     @property
   ...     def size(self):
   ...         return 161
   ...
   ...     def get_capacity(self):
   ...         return super().get_capacity()

   >>> bikeshed = BikeShed()
   >>> bikeshed.get_capacity()
   23
""",
    'maintainer': 'Ryan Northey',
    'maintainer_email': 'ryan@synca.io',
    'name': 'abstracts',
    'namespace_packages': (
    ),
    'package_data': {
        'abstracts': (
            'py.typed',
        ),
    },
    'packages': (
        'abstracts',
    ),
    'python_requires': '<3.13,>=3.12',
    'url': 'https://github.com/envoyproxy/toolshed/tree/main/py/abstracts',
    'version': '0.0.23',
})



from .implements import Implementer
from .interface import Interface
from .decorators import implementer, interfacemethod
from .abstraction import Abstraction


__all__ = (
    "Abstraction",
    "Interface",
    "implementer",
    "interfacemethod",
    "Implementer")
